@@ -666,7 +666,8 @@ class WindowClient(StubClientMixin):
         else:
             from xpra.codecs.pillow.decoder import open_only
             img = open_only(data, ("png", ))
-            assert img.mode in ("RGB", "RGBA"), "invalid image mode: %s" % img.mode
+            if img.mode not in ("RGB", "RGBA"):
+                img = img.convert("RGBA")
             has_alpha = img.mode=="RGBA"
             rowstride = width * (3+int(has_alpha))
         icon = img
@@ -878,7 +879,9 @@ class WindowClient(StubClientMixin):
     def deiconify_windows(self):
         log("deiconify_windows()")
         for window in self._id_to_window.values():
-            window.deiconify()
+            deiconify = getattr(window, "deiconify", None)
+            if deiconify:
+                deiconify()
 
 
     def resize_windows(self, new_size_fn):
